#===============================================================================
#
#     Neurons.py
#
#     This file is part of ANNarchy.
#
#     Copyright (C) 2013-2016  Julien Vitay <julien.vitay@gmail.com>,
#     Helge Uelo Dinkelbach <helge.dinkelbach@gmail.com>
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     ANNarchy is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#===============================================================================
from ANNarchy.core.Neuron import Neuron

def list_standard_neurons():
    "Returns a list of standard neuron models available."
    return [LeakyIntegrator, Izhikevich, IF_curr_exp, IF_cond_exp, IF_curr_alpha, IF_cond_alpha, HH_cond_exp, EIF_cond_alpha_isfa_ista, EIF_cond_exp_isfa_ista]


##################
### Leaky Integrator
##################
class LeakyIntegrator(Neuron):
    '''
    Leaky-integrator rate-coded neuron, optionally noisy.

    This simple rate-coded neuron defines an internal variable :math:`v(t)` which integrates the inputs :math:`I(t)` with a time constant :math:`\tau` and a baseline ``B``. An additive noise :math:`N(t)` can be optionally defined: 

    .. math::

        \\tau \cdot \\frac{dv(t)}{dt} + v(t) = I(t) + B + N(t)

    The transfer function is the positive (or rectified linear ReLU) function with a thresholf :math:`T`:

    .. math::

        r(t) = (v(t) - T)^+

    By default, the input ``I(t)`` to this neuron is "sum(exc) - sum(inh)", but this can be changed by setting the ``sum`` argument::

        neuron = LeakyIntegrator(sum="sum('exc')")

    By default, there is no additive noise, but the ``noise`` argument can be passed with a specific distribution::

        neuron = LeakyIntegrator(noise="Normal(0.0, 1.0)")


    Parameters:

    * tau = 10.0 : Time constant in ms of the neuron.
    * B = 0.0 : Baseline value for v.
    * T = 0.0 : Threshold for the positive transfer function.

    Variables:

    * v : internal variable (init = 0.0)::

        tau * dv/dt + v = sum(exc) - sum(inh) + B + N

    * r : firing rate (init = 0.0)::

        r = pos(v - T)

    The ODE is solved using the exponential Euler method.

    Equivalent code:

    .. code-block:: python

        LeakyIntegrator = Neuron(
            parameters="""
                tau = 10.0 : population
                B = 0.0
                T = 0.0 : population
            """, 
            equations="""
                tau * dv/dt + v = sum(exc) - sum(inh) + B : exponential
                r = pos(v - T)
            """
        )

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, tau=10.0, B=0.0, T=0.0, sum='sum(exc) - sum(inh)', noise=None):
        
        # Create the arguments
        parameters = """
    tau = %(tau)s : population
    B = %(B)s
    T = %(T)s : population
""" % {'tau': tau, 'B': B, 'T': T}

        # Equations for the variables
        if not noise:
            noise_def = ''
        else:
            noise_def = '+ ' + noise

        equations="""
    tau * dv/dt + v = %(sum)s + B %(noise)s : exponential
    r = pos(v - T)
""" % { 'sum' : sum, 'noise': noise_def}

        Neuron.__init__(self, parameters=parameters, equations=equations,
            name="Leaky-Integrator", description="Leaky-Integrator with positive transfer function and additive noise.")

        # For reporting
        self._instantiated.append(True)


##################
### Izhikevich
##################
class Izhikevich(Neuron):
    '''
    Izhikevich neuron as proposed in

    Izhikevich, E.M. (2003). *Simple Model of Spiking Neurons, IEEE Transaction on Neural Networks*, 14:6. `doi:10.1109/TNN.2003.820440 <http://dx.doi.org/10.1109/TNN.2003.820440>`_

    The equations are

    .. math::

        \\frac{dv}{dt} = 0.04 * v^2 + 5.0 * v + 140.0 - u + I 

        \\frac{du}{dt} = a * (b * v - u) 

    By default, the conductance is "g_exc - g_inh", but this can be changed by setting the ``conductance`` argument::

        neuron = Izhikevich(conductance='g_ampa * (1 + g_nmda) - g_gaba')

    The synapses are instantaneous, i.e the corresponding conductance is increased from the synaptic efficiency w at the time step when a spike is received.

    Parameters:

    * a = 0.02 : Speed of the recovery variable
    * b = 0.2: Scaling of the recovery variable
    * c = -65.0 : Reset potential.
    * d = 8.0 : Increment of the recovery variable after a spike.
    * v_thresh = 30.0 : Spike threshold (mV).
    * i_offset = 0.0 : external current (nA).
    * noise = 0.0 : Amplitude of the normal additive noise.
    * tau_refrac = 0.0 : Duration of refractory period (ms).

    Variables:

    * I : input current (user-defined conductance/current + external current + normal noise)::

        I = conductance + i_offset + noise * Normal(0.0, 1.0)

    * v : membrane potential in mV (init = c)::

        dv/dt = 0.04 * v^2 + 5.0 * v + 140.0 - u + I 

    * u : recovery variable (init= b * c)::

        du/dt = a * (b * v - u) 

    Spike emission::

        v > v_thresh

    Reset::

        v = c
        u += d 

    The ODEs are solved using the explicit Euler method.

    Equivalent code:

    .. code-block:: python

        Izhikevich = Neuron(
            parameters = """
                noise = 0.0
                a = 0.02
                b = 0.2
                c = -65.0
                d = 8.0
                v_thresh = 30.0
                i_offset = 0.0
            """, 
            equations = """
                I = g_exc - g_inh + noise * Normal(0.0, 1.0) + i_offset
                dv/dt = 0.04 * v^2 + 5.0 * v + 140.0 - u + I : init = -65.0
                du/dt = a * (b*v - u) : init= -13.0
            """,
            spike = "v > v_thresh",
            reset = "v = c; u += d",
            refractory = 0.0
        )

    The default parameters are for a regular spiking (RS) neuron derived from the above mentioned article.
    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, a=0.02, b=0.2, c=-65.0, d=8.0, v_thresh=30.0, i_offset=0.0, noise=0.0, tau_refrac=0.0, conductance="g_exc - g_inh"):
        # Extract which targets are defined in the conductance
        import re
        targets = re.findall(r'g_([\w]+)', conductance)
        # Create the arguments
        parameters = """
    noise = %(noise)s
    a = %(a)s
    b = %(b)s
    c = %(c)s
    d = %(d)s
    v_thresh = %(v_thresh)s
    i_offset = %(i_offset)s
    tau_refrac = %(tau_refrac)s
""" % {'a': a, 'b':b, 'c':c, 'd':d, 'v_thresh':v_thresh, 'i_offset':i_offset, 'noise':noise, 'tau_refrac':tau_refrac}
        # Equations for the variables
        equations="""
    I = %(conductance)s + noise * Normal(0.0, 1.0) + i_offset
    dv/dt = 0.04 * v^2 + 5.0 * v + 140.0 - u + I : init = %(c)s
    du/dt = a * (b*v - u) : init= %(u)s
""" % { 'conductance' : conductance, 'c':c , 'u': b*c}

    #     # Default behavior for the conductances (avoid warning)
    #     for target in targets:
    #         equations += """
    # g_%(target)s = 0.0""" % {'target' : target}

        spike = """
    v > v_thresh
"""
        reset = """
    v = c
    u += d
"""
        Neuron.__init__(self, parameters=parameters, equations=equations, spike=spike, reset=reset, refractory='tau_refrac',
            name="Izhikevich", description="Quadratic integrate-and-fire spiking neuron with adaptation.")

        # For reporting
        self._instantiated.append(True)


##################
### IF neurons
##################
class IF_curr_exp(Neuron):
    '''
    IF_curr_exp neuron.

    Leaky integrate-and-fire model with fixed threshold and decaying-exponential post-synaptic current. (Separate synaptic currents for excitatory and inhibitory synapses).

    Parameters:

    * v_rest = -65.0 :  Resting membrane potential (mV)
    * cm  = 1.0 : Capacity of the membrane (nF)
    * tau_m  = 20.0 : Membrane time constant (ms)
    * tau_refrac = 0.0 : Duration of refractory period (ms)
    * tau_syn_E = 5.0 : Decay time of excitatory synaptic current (ms)
    * tau_syn_I = 5.0 : Decay time of inhibitory synaptic current (ms)
    * i_offset = 0.0 : Offset current (nA)
    * v_reset = -65.0 : Reset potential after a spike (mV)
    * v_thresh = -50.0 : Spike threshold (mV)

    Variables:

    * v : membrane potential in mV (init=-65.0)::

        cm * dv/dt = cm/tau_m*(v_rest -v)   + g_exc - g_inh + i_offset

    * g_exc : excitatory current (init = 0.0)::

        tau_syn_E * dg_exc/dt = - g_exc

    * g_inh : inhibitory current (init = 0.0)::

        tau_syn_I * dg_inh/dt = - g_inh


    Spike emission::

        v > v_thresh

    Reset::

        v = v_reset

    The ODEs are solved using the exponential Euler method.
    
    Equivalent code:

    .. code-block:: python

        IF_curr_exp = Neuron(
            parameters = """
                v_rest = -65.0
                cm  = 1.0
                tau_m  = 20.0
                tau_syn_E = 5.0
                tau_syn_I = 5.0
                v_thresh = -50.0
                v_reset = -65.0
                i_offset = 0.0
            """, 
            equations = """
                cm * dv/dt = cm/tau_m*(v_rest -v)   + g_exc - g_inh + i_offset : exponential, init=-65.0
                tau_syn_E * dg_exc/dt = - g_exc : exponential
                tau_syn_I * dg_inh/dt = - g_inh : exponential
            """,
            spike = "v > v_thresh",
            reset = "v = v_reset",
            refractory = 0.0
        )

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest=-65.0, cm=1.0, tau_m=20.0, tau_refrac=0.0, tau_syn_E=5.0, tau_syn_I=5.0, v_thresh=-50.0, v_reset=-65.0, i_offset=0.0):
        # Create the arguments
        parameters = """
    v_rest = %(v_rest)s
    cm  = %(cm)s
    tau_m  = %(tau_m)s
    tau_refrac = %(tau_refrac)s
    tau_syn_E = %(tau_syn_E)s
    tau_syn_I = %(tau_syn_I)s
    v_thresh = %(v_thresh)s
    v_reset = %(v_reset)s
    i_offset = %(i_offset)s
""" % {'v_rest':v_rest, 'cm':cm, 'tau_m':tau_m, 'tau_refrac':tau_refrac, 'tau_syn_E':tau_syn_E, 'tau_syn_I':tau_syn_I, 'v_thresh':v_thresh, 'v_reset':v_reset, 'i_offset':i_offset}
        # Equations for the variables
        equations="""    
    cm * dv/dt = cm/tau_m*(v_rest -v)   + g_exc - g_inh + i_offset : exponential, init=%(v_reset)s
    tau_syn_E * dg_exc/dt = - g_exc : exponential
    tau_syn_I * dg_inh/dt = - g_inh : exponential
""" % {'v_reset':v_reset}

        spike = """
    v > v_thresh
"""
        reset = """
    v = v_reset
"""
        Neuron.__init__(self, parameters=parameters, equations=equations, spike=spike, reset=reset, refractory='tau_refrac',
            name="Integrate-and-Fire", description="Leaky integrate-and-fire model with fixed threshold and decaying-exponential post-synaptic current.")

        # For reporting
        self._instantiated.append(True)

class IF_cond_exp(Neuron):
    '''
    IF_cond_exp neuron.

    Leaky integrate-and-fire model with fixed threshold and decaying-exponential post-synaptic conductance.

    Parameters:

    * v_rest = -65.0 :  Resting membrane potential (mV)
    * cm  = 1.0 : Capacity of the membrane (nF)
    * tau_m  = 20.0 : Membrane time constant (ms)
    * tau_refrac = 0.0 : Duration of refractory period (ms)
    * tau_syn_E = 5.0 : Decay time of excitatory synaptic current (ms)
    * tau_syn_I = 5.0 : Decay time of inhibitory synaptic current (ms)
    * e_rev_E = 0.0 : Reversal potential for excitatory input (mV)
    * e_rev_I = -70.0 : Reversal potential for inhibitory input (mv)
    * i_offset = 0.0 : Offset current (nA)
    * v_reset = -65.0 : Reset potential after a spike (mV)
    * v_thresh = -50.0 : Spike threshold (mV)

    Variables:

    * v : membrane potential in mV (init=-65.0)::

        cm * dv/dt = cm/tau_m*(v_rest -v)  + g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset

    * g_exc : excitatory current (init = 0.0)::

        tau_syn_E * dg_exc/dt = - g_exc

    * g_inh : inhibitory current (init = 0.0)::

        tau_syn_I * dg_inh/dt = - g_inh


    Spike emission::

        v > v_thresh

    Reset::

        v = v_reset

    The ODEs are solved using the exponential Euler method.

    Equivalent code:

    .. code-block:: python

        IF_cond_exp = Neuron(
            parameters = """
                v_rest = -65.0
                cm  = 1.0
                tau_m  = 20.0
                tau_syn_E = 5.0
                tau_syn_I = 5.0
                e_rev_E = 0.0
                e_rev_I = -70.0
                v_thresh = -50.0
                v_reset = -65.0
                i_offset = 0.0
            """, 
            equations = """
                cm * dv/dt = cm/tau_m*(v_rest -v)   + g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset : exponential, init=-65.0
                tau_syn_E * dg_exc/dt = - g_exc : exponential
                tau_syn_I * dg_inh/dt = - g_inh : exponential
            """,
            spike = "v > v_thresh",
            reset = "v = v_reset",
            refractory = 0.0
        )

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest=-65.0, cm=1.0, tau_m=20.0, tau_refrac=0.0, tau_syn_E=5.0, tau_syn_I=5.0, e_rev_E = 0.0, e_rev_I = -70.0, v_thresh=-50.0, v_reset=-65.0, i_offset=0.0):
        # Create the arguments
        parameters = """
    v_rest = %(v_rest)s
    cm  = %(cm)s
    tau_m  = %(tau_m)s
    tau_refrac = %(tau_refrac)s
    tau_syn_E = %(tau_syn_E)s
    tau_syn_I = %(tau_syn_I)s
    v_thresh = %(v_thresh)s
    v_reset = %(v_reset)s
    i_offset = %(i_offset)s
    e_rev_E = %(e_rev_E)s 
    e_rev_I = %(e_rev_I)s
""" % {'v_rest':v_rest, 'cm':cm, 'tau_m':tau_m, 'tau_refrac':tau_refrac, 'tau_syn_E':tau_syn_E, 'tau_syn_I':tau_syn_I, 'v_thresh':v_thresh, 'v_reset':v_reset, 'i_offset':i_offset, 'e_rev_E': e_rev_E, 'e_rev_I': e_rev_I}
        # Equations for the variables
        equations="""    
    cm * dv/dt = cm/tau_m*(v_rest -v)   + g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset : exponential, init=%(v_reset)s
    tau_syn_E * dg_exc/dt = - g_exc : exponential
    tau_syn_I * dg_inh/dt = - g_inh : exponential
"""% {'v_reset':v_reset}

        spike = """
    v > v_thresh
"""
        reset = """
    v = v_reset
"""
        Neuron.__init__(self, parameters=parameters, equations=equations, spike=spike, reset=reset, refractory='tau_refrac',
            name="Integrate-and-Fire", description="Leaky integrate-and-fire model with fixed threshold and decaying-exponential post-synaptic conductances.")

        # For reporting
        self._instantiated.append(True)

# Alpha conductances
class IF_curr_alpha(Neuron):
    '''
    IF_curr_alpha neuron.

    Leaky integrate-and-fire model with fixed threshold and alpha post-synaptic currents. (Separate synaptic currents for excitatory and inhibitory synapses).

    The alpha currents are calculated through a system of two linears ODEs. After a spike is received at t_spike, it peaks at t_spike + tau_syn_X, with a maximum equal to the synaptic efficiency.

    Parameters:

    * v_rest = -65.0 :  Resting membrane potential (mV)
    * cm  = 1.0 : Capacity of the membrane (nF)
    * tau_m  = 20.0 : Membrane time constant (ms)
    * tau_refrac = 0.0 : Duration of refractory period (ms)
    * tau_syn_E = 5.0 : Rise time of excitatory synaptic current (ms)
    * tau_syn_I = 5.0 : Rise time of inhibitory synaptic current (ms)
    * i_offset = 0.0 : Offset current (nA)
    * v_reset = -65.0 : Reset potential after a spike (mV)
    * v_thresh = -50.0 : Spike threshold (mV)

    Variables:

    * v : membrane potential in mV (init=-65.0)::

        cm * dv/dt = cm/tau_m*(v_rest -v) + alpha_exc - alpha_inh + i_offset

    * g_exc : excitatory current (init = 0.0)::

        tau_syn_E * dg_exc/dt = - g_exc

    * alpha_exc : alpha function of excitatory current (init = 0.0)::

        tau_syn_E * dalpha_exc/dt = exp((tau_syn_E - dt/2.0)/tau_syn_E) * g_exc - alpha_exc

    * g_inh : inhibitory current (init = 0.0)::

        tau_syn_I * dg_inh/dt = - g_inh

    * alpha_inh : alpha function of inhibitory current (init = 0.0)::

        tau_syn_I * dalpha_inh/dt = exp((tau_syn_I - dt/2.0)/tau_syn_I) * g_inh - alpha_inh


    Spike emission::

        v > v_thresh

    Reset::

        v = v_reset

    The ODEs are solved using the exponential Euler method.

    Equivalent code:

    .. code-block:: python

        IF_curr_alpha = Neuron(
            parameters = """
                v_rest = -65.0
                cm  = 1.0
                tau_m  = 20.0
                tau_syn_E = 5.0
                tau_syn_I = 5.0
                v_thresh = -50.0
                v_reset = -65.0
                i_offset = 0.0
            """, 
            equations = """
                gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)
                gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)  
                cm * dv/dt = cm/tau_m*(v_rest -v)   + alpha_exc - alpha_inh + i_offset : exponential, init=-65.0
                tau_syn_E * dg_exc/dt = - g_exc : exponential
                tau_syn_I * dg_inh/dt = - g_inh : exponential
                tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc  : exponential
                tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh  : exponential
            """,
            spike = "v > v_thresh",
            reset = "v = v_reset",
            refractory = 0.0
        )

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest=-65.0, cm=1.0, tau_m=20.0, tau_refrac=0.0, tau_syn_E=5.0, tau_syn_I=5.0, v_thresh=-50.0, v_reset=-65.0, i_offset=0.0):
        # Create the arguments
        parameters = """
    v_rest = %(v_rest)s
    cm  = %(cm)s
    tau_m  = %(tau_m)s
    tau_refrac = %(tau_refrac)s
    tau_syn_E = %(tau_syn_E)s
    tau_syn_I = %(tau_syn_I)s
    v_thresh = %(v_thresh)s
    v_reset = %(v_reset)s
    i_offset = %(i_offset)s
""" % {'v_rest':v_rest, 'cm':cm, 'tau_m':tau_m, 'tau_refrac':tau_refrac, 'tau_syn_E':tau_syn_E, 'tau_syn_I':tau_syn_I, 'v_thresh':v_thresh, 'v_reset':v_reset, 'i_offset':i_offset}
        # Equations for the variables
        equations="""  
    gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)
    gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)  
    cm * dv/dt = cm/tau_m*(v_rest -v)   + alpha_exc - alpha_inh + i_offset : exponential, init=%(v_reset)s
    tau_syn_E * dg_exc/dt = - g_exc : exponential
    tau_syn_I * dg_inh/dt = - g_inh : exponential
    tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc  : exponential
    tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh  : exponential
"""  % {'v_reset':v_reset}

        spike = """
    v > v_thresh
"""
        reset = """
    v = v_reset
"""
        Neuron.__init__(self, parameters=parameters, equations=equations, spike=spike, reset=reset, refractory='tau_refrac',
            name="Integrate-and-Fire", description="Leaky integrate-and-fire model with fixed threshold and alpha post-synaptic currents.")

        # For reporting
        self._instantiated.append(True)


class IF_cond_alpha(Neuron):
    '''
    IF_cond_exp neuron.

    Leaky integrate-and-fire model with fixed threshold and alpha post-synaptic conductance.

    Parameters:

    * v_rest = -65.0 :  Resting membrane potential (mV)
    * cm  = 1.0 : Capacity of the membrane (nF)
    * tau_m  = 20.0 : Membrane time constant (ms)
    * tau_refrac = 0.0 : Duration of refractory period (ms)
    * tau_syn_E = 5.0 : Rise time of excitatory synaptic current (ms)
    * tau_syn_I = 5.0 : Rise time of inhibitory synaptic current (ms)
    * e_rev_E = 0.0 : Reversal potential for excitatory input (mV)
    * e_rev_I = -70.0 : Reversal potential for inhibitory input (mv)
    * i_offset = 0.0 : Offset current (nA)
    * v_reset = -65.0 : Reset potential after a spike (mV)
    * v_thresh = -50.0 : Spike threshold (mV)

    Variables:

    * v : membrane potential in mV (init=-65.0)::

        cm * dv/dt = cm/tau_m*(v_rest -v)  + alpha_exc * (e_rev_E - v) + alpha_inh * (e_rev_I - v) + i_offset

    * g_exc : excitatory conductance (init = 0.0)::

        tau_syn_E * dg_exc/dt = - g_exc

    * alpha_exc : alpha function of excitatory conductance (init = 0.0)::

        tau_syn_E * dalpha_exc/dt = exp((tau_syn_E - dt/2.0)/tau_syn_E) * g_exc - alpha_exc

    * g_inh : inhibitory conductance (init = 0.0)::

        tau_syn_I * dg_inh/dt = - g_inh

    * alpha_inh : alpha function of inhibitory current (init = 0.0)::

        tau_syn_I * dalpha_inh/dt = exp((tau_syn_I - dt/2.0)/tau_syn_I) * g_inh - alpha_inh


    Spike emission::

        v > v_thresh

    Reset::

        v = v_reset

    The ODEs are solved using the exponential Euler method.

    Equivalent code:

    .. code-block:: python

        IF_cond_alpha = Neuron(
            parameters = """
                v_rest = -65.0
                cm  = 1.0
                tau_m  = 20.0
                tau_syn_E = 5.0
                tau_syn_I = 5.0
                e_rev_E = 0.0
                e_rev_I = -70.0
                v_thresh = -50.0
                v_reset = -65.0
                i_offset = 0.0
            """, 
            equations = """
                gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)
                gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)
                cm * dv/dt = cm/tau_m*(v_rest -v)   + alpha_exc * (e_rev_E - v) + alpha_inh * (e_rev_I - v) + i_offset : exponential, init=-65.0
                tau_syn_E * dg_exc/dt = - g_exc : exponential
                tau_syn_I * dg_inh/dt = - g_inh : exponential
                tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc  : exponential
                tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh  : exponential
            """,
            spike = "v > v_thresh",
            reset = "v = v_reset",
            refractory = 0.0
        )

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest=-65.0, cm=1.0, tau_m=20.0, tau_refrac=0.0, tau_syn_E=5.0, tau_syn_I=5.0, e_rev_E = 0.0, e_rev_I = -70.0, v_thresh=-50.0, v_reset=-65.0, i_offset=0.0):
        # Create the arguments
        parameters = """
    v_rest = %(v_rest)s
    cm  = %(cm)s
    tau_m  = %(tau_m)s
    tau_refrac = %(tau_refrac)s
    tau_syn_E = %(tau_syn_E)s
    tau_syn_I = %(tau_syn_I)s
    v_thresh = %(v_thresh)s
    v_reset = %(v_reset)s
    i_offset = %(i_offset)s
    e_rev_E = %(e_rev_E)s 
    e_rev_I = %(e_rev_I)s
""" % {'v_rest':v_rest, 'cm':cm, 'tau_m':tau_m, 'tau_refrac':tau_refrac, 'tau_syn_E':tau_syn_E, 'tau_syn_I':tau_syn_I, 'v_thresh':v_thresh, 'v_reset':v_reset, 'i_offset':i_offset, 'e_rev_E': e_rev_E, 'e_rev_I': e_rev_I}
        # Equations for the variables
        equations="""    
    gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)
    gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)
    cm * dv/dt = cm/tau_m*(v_rest -v)   + alpha_exc * (e_rev_E - v) + alpha_inh * (e_rev_I - v) + i_offset : exponential, init=%(v_reset)s
    tau_syn_E * dg_exc/dt = - g_exc : exponential
    tau_syn_I * dg_inh/dt = - g_inh : exponential
    tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc  : exponential
    tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh  : exponential
""" % {'v_reset': v_reset}

        spike = """
    v > v_thresh
"""
        reset = """
    v = v_reset
"""
        Neuron.__init__(self, parameters=parameters, equations=equations, spike=spike, reset=reset, refractory='tau_refrac',
            name="Integrate-and-Fire", description="Leaky integrate-and-fire model with fixed threshold and alpha post-synaptic conductances.")

        # For reporting
        self._instantiated.append(True)

##################
### EIF neurons
##################

class EIF_cond_exp_isfa_ista(Neuron):
    '''
    EIF_cond_exp neuron.

    Exponential integrate-and-fire neuron with spike triggered and sub-threshold adaptation currents (isfa, ista reps.) according to:

    Brette R and Gerstner W (2005) Adaptive Exponential Integrate-and-Fire Model as an Effective Description of Neuronal Activity. J Neurophysiol 94:3637-3642

    Parameters:

    * v_rest = -70.6 :  Resting membrane potential (mV)
    * cm = 0.281 : Capacity of the membrane (nF)
    * tau_m = 9.3667 : Membrane time constant (ms)
    * tau_refrac = 0.1 : Duration of refractory period (ms)
    * tau_syn_E = 5.0 : Decay time of excitatory synaptic current (ms)
    * tau_syn_I = 5.0 : Decay time of inhibitory synaptic current (ms)
    * e_rev_E = 0.0 : Reversal potential for excitatory input (mV)
    * e_rev_I = -80.0 : Reversal potential for inhibitory input (mv)
    * tau_w = 144.0 : Time constant of the adaptation variable (ms)
    * a = 4.0 : Scaling of the adaptation variable
    * b = 0.0805 : Increment on the adaptation variable after a spike
    * i_offset = 0.0 : Offset current (nA)
    * delta_T = 2.0 : Speed of the exponential (mV)
    * v_thresh = -50.4 : Spike threshold for the exponential (mV)
    * v_reset = -70.6 : Reset potential after a spike (mV)
    * v_spike = -40.0 : Spike threshold (mV)

    Variables:

    * I : input current (nA)::

        I = g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset

    * v : membrane potential in mV (init=-70.6)::

        tau_m * dv/dt = (v_rest - v +  delta_T * exp((v-v_thresh)/delta_T)) + tau_m/cm*(I - w)

    * w : adaptation variable (init=0.0)::

        tau_w * dw/dt = a * (v - v_rest) / 1000.0 - w

    * g_exc : excitatory current (init = 0.0)::

        tau_syn_E * dg_exc/dt = - g_exc

    * g_inh : inhibitory current (init = 0.0)::

        tau_syn_I * dg_inh/dt = - g_inh


    Spike emission::

        v > v_thresh

    Reset::

        v = v_reset
        u += b

    The ODEs are solved using the explicit Euler method.

    Equivalent code:

    .. code-block:: python

        EIF_cond_exp_isfa_ista = Neuron(
            parameters = """
                v_rest = -70.6
                cm = 0.281 
                tau_m = 9.3667 
                tau_syn_E = 5.0 
                tau_syn_I = 5.0 
                e_rev_E = 0.0 
                e_rev_I = -80.0
                tau_w = 144.0 
                a = 4.0
                b = 0.0805
                i_offset = 0.0
                delta_T = 2.0 
                v_thresh = -50.4
                v_reset = -70.6
                v_spike = -40.0 
            """, 
            equations = """
                I = g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset            
                tau_m * dv/dt = (v_rest - v +  delta_T * exp((v-v_thresh)/delta_T)) + tau_m/cm*(I - w) : init=-70.6          
                tau_w * dw/dt = a * (v - v_rest) / 1000.0 - w           
                tau_syn_E * dg_exc/dt = - g_exc : exponential
                tau_syn_I * dg_inh/dt = - g_inh : exponential
            """,
            spike = "v > v_spike",
            reset = """
                v = v_reset
                w += b
            """,
            refractory = 0.1
        )

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest = -70.6, cm = 0.281, tau_m = 9.3667, tau_refrac = 0.1, tau_syn_E = 5.0, tau_syn_I = 5.0, e_rev_E = 0.0, e_rev_I = -80.0, tau_w = 144.0, a = 4.0, b = 0.0805, i_offset = 0.0, delta_T = 2.0, v_thresh = -50.4, v_reset = -70.6, v_spike = -40.0):
        # Create the arguments
        parameters = """
    v_rest     = %(v_rest)s
    cm         = %(cm)s
    tau_m      = %(tau_m)s
    tau_refrac = %(tau_refrac)s
    tau_syn_E  = %(tau_syn_E)s
    tau_syn_I  = %(tau_syn_I)s
    e_rev_E    = %(e_rev_E)s
    e_rev_I    = %(e_rev_I)s
    tau_w      = %(tau_w)s
    a          = %(a)s
    b          = %(b)s
    i_offset   = %(i_offset)s
    delta_T    = %(delta_T)s
    v_thresh   = %(v_thresh)s
    v_reset    = %(v_reset)s
    v_spike    = %(v_spike)s
""" % {
    'v_rest'     : v_rest    ,
    'cm'         : cm        ,
    'tau_m'      : tau_m     ,
    'tau_refrac' : tau_refrac,
    'tau_syn_E'  : tau_syn_E ,
    'tau_syn_I'  : tau_syn_I ,
    'e_rev_E'    : e_rev_E   ,
    'e_rev_I'    : e_rev_I   ,
    'tau_w'      : tau_w     ,
    'a'          : a         ,
    'b'          : b         ,
    'i_offset'   : i_offset  ,
    'delta_T'    : delta_T   ,
    'v_thresh'   : v_thresh  ,
    'v_reset'    : v_reset   ,
    'v_spike'    : v_spike   ,
}
        # Equations for the variables
        equations="""    
    I = g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset

    tau_m * dv/dt = (v_rest - v +  delta_T * exp((v-v_thresh)/delta_T)) + tau_m/cm*(I - w) : init=%(v_reset)s

    tau_w * dw/dt = a * (v - v_rest) / 1000.0 - w 

    tau_syn_E * dg_exc/dt = - g_exc : exponential
    tau_syn_I * dg_inh/dt = - g_inh : exponential
""" % {'v_reset': v_reset}

        spike = """
    v > v_spike
"""
        reset = """
    v = v_reset
    w += b
"""
        Neuron.__init__(self, parameters=parameters, equations=equations, spike=spike, reset=reset, refractory='tau_refrac',
            name="Adaptive exponential Integrate-and-Fire", description="Exponential integrate-and-fire neuron with spike triggered and sub-threshold adaptation currents (isfa, ista reps.).")

        # For reporting
        self._instantiated.append(True)

class EIF_cond_alpha_isfa_ista(Neuron):
    ''' 
    EIF_cond_alpha neuron.

    Exponential integrate-and-fire neuron with spike triggered and sub-threshold adaptation conductances (isfa, ista reps.) according to:

    Brette R and Gerstner W (2005) Adaptive Exponential Integrate-and-Fire Model as an Effective Description of Neuronal Activity. J Neurophysiol 94:3637-3642

    Parameters:

    * v_rest = -70.6 :  Resting membrane potential (mV)
    * cm = 0.281 : Capacity of the membrane (nF)
    * tau_m = 9.3667 : Membrane time constant (ms)
    * tau_refrac = 0.1 : Duration of refractory period (ms)
    * tau_syn_E = 5.0 : Decay time of excitatory synaptic current (ms)
    * tau_syn_I = 5.0 : Decay time of inhibitory synaptic current (ms)
    * e_rev_E = 0.0 : Reversal potential for excitatory input (mV)
    * e_rev_I = -80.0 : Reversal potential for inhibitory input (mv)
    * tau_w = 144.0 : Time constant of the adaptation variable (ms)
    * a = 4.0 : Scaling of the adaptation variable
    * b = 0.0805 : Increment on the adaptation variable after a spike
    * i_offset = 0.0 : Offset current (nA)
    * delta_T = 2.0 : Speed of the exponential (mV)
    * v_thresh = -50.4 : Spike threshold for the exponential (mV)
    * v_reset = -70.6 : Reset potential after a spike (mV)
    * v_spike = -40.0 : Spike threshold (mV)

    Variables:

    * I : input current (nA)::

        I = g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset

    * v : membrane potential in mV (init=-70.6)::

        tau_m * dv/dt = (v_rest - v +  delta_T * exp((v-v_thresh)/delta_T)) + tau_m/cm*(I - w)

    * w : adaptation variable (init=0.0)::

        tau_w * dw/dt = a * (v - v_rest) / 1000.0 - w

    * g_exc : excitatory current (init = 0.0)::

        tau_syn_E * dg_exc/dt = - g_exc

    * g_inh : inhibitory current (init = 0.0)::

        tau_syn_I * dg_inh/dt = - g_inh

    * alpha_exc : alpha function of excitatory current (init = 0.0)::

        tau_syn_E * dalpha_exc/dt = exp((tau_syn_E - dt/2.0)/tau_syn_E) * g_exc - alpha_exc  
    
    * alpha_inh: alpha function of inhibitory current (init = 0.0)::

        tau_syn_I * dalpha_inh/dt = exp((tau_syn_I - dt/2.0)/tau_syn_I) * g_inh - alpha_inh  


    Spike emission::

        v > v_spike

    Reset::

        v = v_reset
        u += b

    The ODEs are solved using the explicit Euler method.

    Equivalent code:

    .. code-block:: python

        EIF_cond_alpha_isfa_ista = Neuron(
            parameters = """
                v_rest = -70.6
                cm = 0.281 
                tau_m = 9.3667 
                tau_syn_E = 5.0 
                tau_syn_I = 5.0 
                e_rev_E = 0.0 
                e_rev_I = -80.0
                tau_w = 144.0 
                a = 4.0
                b = 0.0805
                i_offset = 0.0
                delta_T = 2.0 
                v_thresh = -50.4
                v_reset = -70.6
                v_spike = -40.0 
            """, 
            equations = """
                gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)
                gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)                
                I = alpha_exc * (e_rev_E - v) + alpha_inh * (e_rev_I - v) + i_offset
                tau_m * dv/dt = (v_rest - v +  delta_T * exp((v-v_thresh)/delta_T)) + tau_m/cm*(I - w) : init=-70.6
                tau_w * dw/dt = a * (v - v_rest) / 1000.0 - w 
                tau_syn_E * dg_exc/dt = - g_exc : exponential
                tau_syn_I * dg_inh/dt = - g_inh : exponential
                tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc  : exponential
                tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh  : exponential
            """,
            spike = "v > v_spike",
            reset = """
                v = v_reset
                w += b
            """,
            refractory = 0.1
        )

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest = -70.6, cm = 0.281, tau_m = 9.3667, tau_refrac = 0.1, tau_syn_E = 5.0, tau_syn_I = 5.0, e_rev_E = 0.0, e_rev_I = -80.0, tau_w = 144.0, a = 4.0, b = 0.0805, i_offset = 0.0, delta_T = 2.0, v_thresh = -50.4, v_reset = -70.6, v_spike = -40.0):
        # Create the arguments
        parameters = """
    v_rest     = %(v_rest)s
    cm         = %(cm)s
    tau_m      = %(tau_m)s
    tau_refrac = %(tau_refrac)s
    tau_syn_E  = %(tau_syn_E)s
    tau_syn_I  = %(tau_syn_I)s
    e_rev_E    = %(e_rev_E)s
    e_rev_I    = %(e_rev_I)s
    tau_w      = %(tau_w)s
    a          = %(a)s
    b          = %(b)s
    i_offset   = %(i_offset)s
    delta_T    = %(delta_T)s
    v_thresh   = %(v_thresh)s
    v_reset    = %(v_reset)s
    v_spike    = %(v_spike)s
""" % {
    'v_rest'     : v_rest    ,
    'cm'         : cm        ,
    'tau_m'      : tau_m     ,
    'tau_refrac' : tau_refrac,
    'tau_syn_E'  : tau_syn_E ,
    'tau_syn_I'  : tau_syn_I ,
    'e_rev_E'    : e_rev_E   ,
    'e_rev_I'    : e_rev_I   ,
    'tau_w'      : tau_w     ,
    'a'          : a         ,
    'b'          : b         ,
    'i_offset'   : i_offset  ,
    'delta_T'    : delta_T   ,
    'v_thresh'   : v_thresh  ,
    'v_reset'    : v_reset   ,
    'v_spike'    : v_spike   ,
}
        # Equations for the variables
        equations="""    

    gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)
    gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)
    
    I = alpha_exc * (e_rev_E - v) + alpha_inh * (e_rev_I - v) + i_offset

    tau_m * dv/dt = (v_rest - v +  delta_T * exp((v-v_thresh)/delta_T)) + tau_m/cm*(I - w) : init=%(v_reset)s

    tau_w * dw/dt = a * (v - v_rest) / 1000.0 - w 

    tau_syn_E * dg_exc/dt = - g_exc : exponential
    tau_syn_I * dg_inh/dt = - g_inh : exponential
    tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc  : exponential
    tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh  : exponential
""" % {'v_reset': v_reset}

        spike = """
    v > v_spike
"""
        reset = """
    v = v_reset
    w += b
"""
        Neuron.__init__(self, parameters=parameters, equations=equations, spike=spike, reset=reset, refractory='tau_refrac',
            name="Adaptive exponential Integrate-and-Fire", description="Exponential integrate-and-fire neuron with spike triggered and sub-threshold adaptation conductances (isfa, ista reps.).")

        # For reporting
        self._instantiated.append(True)

##################
### HH
##################
class HH_cond_exp(Neuron):
    '''
    HH_cond_exp neuron.

    Single-compartment Hodgkin-Huxley-type neuron with transient sodium and delayed-rectifier potassium currents using the ion channel models from Traub.

    Parameters:

    * gbar_Na = 20.0 : Maximal conductance of the Sodium current.
    * gbar_K = 6.0 : Maximal conductance of the Potassium current. 
    * gleak = 0.01 : Conductance of the leak current (nF)  
    * cm = 0.2 : Capacity of the membrane (nF)
    * v_offset = -63.0 :  Threshold for the rate constants (mV)  
    * e_rev_Na = 50.0 : Reversal potential for the Sodium current (mV) 
    * e_rev_K = -90.0 : Reversal potential for the Potassium current (mV)  
    * e_rev_leak = -65.0 : Reversal potential for the leak current (mV)   
    * e_rev_E = 0.0 : Reversal potential for excitatory input (mV)  
    * e_rev_I = -80.0 : Reversal potential for inhibitory input (mV)  
    * tau_syn_E = 0.2 : Decay time of excitatory synaptic current (ms)  
    * tau_syn_I = 2.0 : Decay time of inhibitory synaptic current (ms)   
    * i_offset = 0.0 : Offset current (nA)
    * v_thresh = 0.0 : Threshold for spike emission

    Variables:

    * Voltage-dependent rate constants an, bn, am, bm, ah, bh::

        an = 0.032 * (15.0 - v + v_offset) / (exp((15.0 - v + v_offset)/5.0) - 1.0)
        am = 0.32  * (13.0 - v + v_offset) / (exp((13.0 - v + v_offset)/4.0) - 1.0)
        ah = 0.128 * exp((17.0 - v + v_offset)/18.0) 

        bn = 0.5   * exp ((10.0 - v + v_offset)/40.0)
        bm = 0.28  * (v - v_offset - 40.0) / (exp((v - v_offset - 40.0)/5.0) - 1.0)
        bh = 4.0/(1.0 + exp (( 10.0 - v + v_offset )) )

    * Activation variables n, m, h (h is initialized to 1.0, n and m to 0.0)::

        dn/dt = an * (1.0 - n) - bn * n 
        dm/dt = am * (1.0 - m) - bm * m 
        dh/dt = ah * (1.0 - h) - bh * h 


    * v : membrane potential in mV (init=-65.0)::

        cm * dv/dt = gleak*(e_rev_leak -v) + gbar_K * n**4 * (e_rev_K - v) + gbar_Na * m**3 * h * (e_rev_Na - v)

                    + g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset

    * g_exc : excitatory conductance (init = 0.0)::

        tau_syn_E * dg_exc/dt = - g_exc

    * g_inh : inhibitory conductance (init = 0.0)::

        tau_syn_I * dg_inh/dt = - g_inh


    Spike emission::

        v > v_thresh and v(t-1) < v_thresh (the spike is emitted only once when v crosses the threshold from below)

    The ODEs for n, m, h and v are solved using the midpoint method, while the conductances g_exc and g_inh are solved using the exponential Euler method.

    Equivalent code:

    .. code-block:: python

        HH_cond_exp = Neuron(
            parameters = """
                gbar_Na = 20.0
                gbar_K = 6.0
                gleak = 0.01
                cm = 0.2 
                v_offset = -63.0 
                e_rev_Na = 50.0
                e_rev_K = -90.0 
                e_rev_leak = -65.0
                e_rev_E = 0.0
                e_rev_I = -80.0 
                tau_syn_E = 0.2
                tau_syn_I = 2.0
                i_offset = 0.0
                v_thresh = 0.0
            """, 
            equations = """
                # Previous membrane potential
                prev_v = v

                # Voltage-dependent rate constants
                an = 0.032 * (15.0 - v + v_offset) / (exp((15.0 - v + v_offset)/5.0) - 1.0)
                am = 0.32  * (13.0 - v + v_offset) / (exp((13.0 - v + v_offset)/4.0) - 1.0)
                ah = 0.128 * exp((17.0 - v + v_offset)/18.0) 

                bn = 0.5   * exp ((10.0 - v + v_offset)/40.0)
                bm = 0.28  * (v - v_offset - 40.0) / (exp((v - v_offset - 40.0)/5.0) - 1.0)
                bh = 4.0/(1.0 + exp (( 10.0 - v + v_offset )) )

                # Activation variables
                dn/dt = an * (1.0 - n) - bn * n : init = 0.0, exponential
                dm/dt = am * (1.0 - m) - bm * m : init = 0.0, exponential
                dh/dt = ah * (1.0 - h) - bh * h : init = 1.0, exponential

                # Membrane equation
                cm * dv/dt = gleak*(e_rev_leak -v) + gbar_K * n**4 * (e_rev_K - v) + gbar_Na * m**3 * h * (e_rev_Na - v)
                                + g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset: exponential, init=-65.0

                # Exponentially-decaying conductances
                tau_syn_E * dg_exc/dt = - g_exc : exponential
                tau_syn_I * dg_inh/dt = - g_inh : exponential
            """,
            spike = "(v > v_thresh) and (prev_v <= v_thresh)",
            reset = ""
        )

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, gbar_Na = 20.0, gbar_K = 6.0, gleak = 0.01, cm = 0.2, v_offset = -63.0, e_rev_Na = 50.0, e_rev_K = -90.0, e_rev_leak = -65.0, e_rev_E = 0.0, e_rev_I = -80.0, tau_syn_E = 0.2, tau_syn_I = 2.0, i_offset = 0.0, v_thresh = 0.0):

        parameters = """
gbar_Na    = %(gbar_Na)s   
gbar_K     = %(gbar_K)s     
gleak      = %(gleak)s      
cm         = %(cm)s        
v_offset   = %(v_offset)s 
e_rev_Na   = %(e_rev_Na )s 
e_rev_K    = %(e_rev_K)s   
e_rev_leak = %(e_rev_leak)s   
e_rev_E    = %(e_rev_E)s    
e_rev_I    = %(e_rev_I)s   
tau_syn_E  = %(tau_syn_E)s 
tau_syn_I  = %(tau_syn_I)s 
i_offset   = %(i_offset)s  
v_thresh   = %(v_thresh)s  
""" % {
'gbar_Na'    : gbar_Na   ,
'gbar_K'     : gbar_K     ,
'gleak'      : gleak      ,
'cm'         : cm        ,
'v_offset'   : v_offset  ,
'e_rev_Na'   : e_rev_Na  ,
'e_rev_K'    : e_rev_K   ,
'e_rev_leak' : e_rev_leak   ,
'e_rev_E'    : e_rev_E    ,
'e_rev_I'    : e_rev_I   ,
'tau_syn_E'  : tau_syn_E ,
'tau_syn_I'  : tau_syn_I ,
'i_offset'   : i_offset  ,
'v_thresh'   : v_thresh  
}

        equations = """
    # Previous membrane potential
    prev_v = v

    # Voltage-dependent rate constants
    an = 0.032 * (15.0 - v + v_offset) / (exp((15.0 - v + v_offset)/5.0) - 1.0)
    am = 0.32  * (13.0 - v + v_offset) / (exp((13.0 - v + v_offset)/4.0) - 1.0)
    ah = 0.128 * exp((17.0 - v + v_offset)/18.0) 

    bn = 0.5   * exp ((10.0 - v + v_offset)/40.0)
    bm = 0.28  * (v - v_offset - 40.0) / (exp((v - v_offset - 40.0)/5.0) - 1.0)
    bh = 4.0/(1.0 + exp (( 10.0 - v + v_offset )) )

    # Activation variables
    dn/dt = an * (1.0 - n) - bn * n : init = 0.0, exponential
    dm/dt = am * (1.0 - m) - bm * m : init = 0.0, exponential
    dh/dt = ah * (1.0 - h) - bh * h : init = 1.0, exponential

    # Membrane equation
    cm * dv/dt = gleak*(e_rev_leak -v) + gbar_K * n**4 * (e_rev_K - v) + gbar_Na * m**3 * h * (e_rev_Na - v)
                    + g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset: exponential, init=%(e_rev_leak)s

    # Exponentially-decaying conductances
    tau_syn_E * dg_exc/dt = - g_exc : exponential
    tau_syn_I * dg_inh/dt = - g_inh : exponential
""" % {'e_rev_leak': e_rev_leak}

        spike = "(v > v_thresh) and (prev_v <= v_thresh)"

        reset = ""

        Neuron.__init__(self, parameters=parameters, equations=equations, spike=spike, reset=reset,
            name="Hodgkin-Huxley", description="Single-compartment Hodgkin-Huxley-type neuron with transient sodium and delayed-rectifier potassium currents.")

        # For reporting
        self._instantiated.append(True)


#   ANNarchy - GapJunction
#
#   A simple network with gap junctions.
# 
#   This is a reimplementation of the Brian example:
#
#   http://brian2.readthedocs.org/en/2.0b3/examples/synapses.gapjunctions.html
#
#   authors: Julien Vitay, Helge Uelo Dinkelbach
from ANNarchy import *

setup(dt=0.1)

neuron = Neuron(
    parameters = "v0 = 1.05: population; tau = 10.0: population",
    equations = "tau*dv/dt = v0 - v + g_gap",
    spike = "v >  1.",
    reset = "v = 0."
)

gap_junction = Synapse(
    psp = "w * (pre.v - post.v)"
)

pop = Population(10, neuron)
pop.v = np.linspace(0., 1., 10)

proj = Projection(pop, pop, 'gap', gap_junction)
proj.connect_all_to_all(0.02)

trace = Monitor(pop[0] + pop[5], 'v')

compile()

simulate(500.)

data = trace.get('v')

import matplotlib.pyplot as plt
plt.plot(data[:, 0])
plt.plot(data[:, 1])
plt.xlabel('Time (ms)')
plt.ylabel('v')
plt.show()


from sympy import Symbol, Function
import ANNarchy.core.Global as Global

# Dictionary of default elements for the C++ generation
parser_dict = {
      'dt' : Symbol('dt'),
      't' : Symbol('(double(t)*dt)'),
      'w' : Symbol('w%(local_index)s'),
      'g_target': Symbol('sum'), # TODO: still useful?
      't_last': Symbol('((double)(last_spike%(local_index)s)*dt)'),
      't_pre': Symbol('((double)(%(pre_prefix)slast_spike[pre_rank%(local_index)s])*dt)'),
      't_post': Symbol('((double)(%(post_prefix)slast_spike[post_rank%(semiglobal_index)s])*dt)'),
}

# Dictionary of built-in functions for the C++ generation
functions_dict = {
      'pos': Function('positive'),
      'positive': Function('positive'),
      'neg': Function('negative'),
      'negative': Function('negative'),
      'modulo': Function('modulo'),
      'fabs': Function('fabs'),
      'ite': Function('ite', nargs=3),
      'clip': Function('clip'),
      'True': Symbol('true'),
      'False': Symbol('false'),
      'power': Function('power', nargs=2),
}

# Built-in functions with their correct name
user_functions = {
    'pos': 'positive',
    'positive': 'positive',
    'neg': 'negative',
    'negative': 'negative',
    'modulo': 'modulo',
    'fabs': 'fabs',
    'clip': 'clip',
    'ite': 'ite',
    'power': 'power'
}

def create_local_dict(local_attributes, semiglobal_attributes, global_attributes, untouched, local_functions=[]):
    """
    Creates a dictionary of Sympy symbols based on the attributes of the neuron/synapse.

    *Parameters:*

    * `local_attributes`: local attributes.
    * `semiglobal_attributes`: semiglobal attributes (postsynaptic).
    * `global_attributes`: global attributes.
    * `untouched`: list of variable names which were replaced (e.g. random distributions, targets)

    """

    # Copy the default dictionary of built-in symbols or functions
    local_dict = parser_dict.copy()
    local_dict.update(functions_dict)

    # Add each variable of the neuron depending on its locality
    for var in local_attributes:
        if var in functions_dict.keys():
            Global._error(var, "is a reserved keyword in ANNarchy.")
        local_dict[var] = Symbol(var + '%(local_index)s')
    for var in semiglobal_attributes:
        if var in functions_dict.keys():
            Global._error(var, "is a reserved keyword in ANNarchy.")
        local_dict[var] = Symbol(var + '%(semiglobal_index)s')
    for var in global_attributes:
        if var in functions_dict.keys():
            Global._error(var, "is a reserved keyword in ANNarchy.")
        local_dict[var] = Symbol(var + '%(global_index)s')

    # Add custom constants
    for obj in Global._objects['constants']:
        if obj.name in local_dict.keys():
            continue
        local_dict[obj.name] = Symbol(obj.name)

    # Add each untouched variable
    for var in untouched:
        local_dict[var] = Symbol(var)

    # Add local functions
    for var in local_functions:
        local_dict[var] = Function(var)

    return local_dict



#===============================================================================
#
#     CSR_OpenmMP.py
#
#     This file is part of ANNarchy.
#
#     Copyright (C) 2016-2018  Julien Vitay <julien.vitay@gmail.com>,
#     Helge Uelo Dinkelbach <helge.dinkelbach@gmail.com>
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     ANNarchy is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#===============================================================================
connectivity_matrix = {
    'declare': """
    // LIL connectivity, just for interface, updated by set_row_ptr()
    std::vector<int> post_ranks;

    // CSR connectivity
    std::vector<int> _row_ptr;
    std::vector<int> _col_idx;
    int _nb_synapses;
""",
    'accessor': """
    // Accessor to connectivity data
    std::vector<int> get_post_rank() { return post_ranks; }
    int nb_synapses(int n) { return _nb_synapses; }

    // LIL specific, read-only
    std::vector<int> get_dendrite_pre_rank(int n) {
        auto beg = _col_idx.begin()+_row_ptr[n];
        auto end = _col_idx.begin()+_row_ptr[n+1];
        return std::vector<int>(beg, end);
    }

    // CSR specific
    void set_row_ptr(std::vector<int> row_ptr) {
        // set new row ptr array
        _row_ptr = row_ptr;
        
        // update the read-only vector post_ranks
        post_ranks.clear();
        for(int i = 0; i < row_ptr.size()-1; i++)
            if ( row_ptr[i] != row_ptr[i+1] )
                post_ranks.push_back(i);
    }
    void set_col_idx(std::vector<int> col_idx) {
        _col_idx = col_idx;
        _nb_synapses = _col_idx.size(); 
    }
""",
    'init': """
""",
    'pyx_struct': """
        # LIL Connectivity, read-only !!!
        vector[int] get_post_rank()
        vector[int] get_dendrite_pre_rank(int)

        # CSR Connectivity
        void set_row_ptr(vector[int])
        void set_col_idx(vector[int])
        void inverse_connectivity_matrix()
""",
    'pyx_wrapper_args': "synapses",
    'pyx_wrapper_init': """
        cdef %(csr_type)s syn = synapses

        proj%(id_proj)s.set_row_ptr(syn._matrix.row_begin())
        proj%(id_proj)s.set_col_idx(syn._matrix.column_indices())
        proj%(id_proj)s.inverse_connectivity_matrix()
""",
    'pyx_wrapper_accessor': """
    # Connectivity
    def post_rank(self):
        return proj%(id_proj)s.get_post_rank()
    def set_post_rank(self, val):
        # Is updated internally
        pass
    def pre_rank(self, int n):
        return proj%(id_proj)s.get_dendrite_pre_rank(n)
"""
}

weight_matrix = {
    'declare': """
    std::vector<%(float_prec)s> w;
""",
    'accessor': """
    void set_w_csr(std::vector<%(float_prec)s> w) { this->w = w; }
    std::vector< %(float_prec)s > get_dendrite_w(int rk) {
        return std::vector<%(float_prec)s>(w.begin()+_row_ptr[rk], w.begin()+_row_ptr[rk+1]);
    }
    std::vector< std::vector<%(float_prec)s> > get_w() {
        std::vector< std::vector<%(float_prec)s> > res;
        for(auto it = post_ranks.begin(); it != post_ranks.end(); it++ ) {
            res.push_back(get_dendrite_w(*it));
        }
        return res;
    }
""",
    'accessor_inv': """
    void set_w_csr(std::vector<%(float_prec)s> w) { this->w = w; }
    std::vector< %(float_prec)s > get_dendrite_w(int rk) {
        std::vector<%(float_prec)s> res;
        for(int j = _col_ptr[rk]; j < _col_ptr[rk+1]; j++)
            res.push_back(w[_inv_idx[j]]);
        return res;
    }
    std::vector< std::vector<%(float_prec)s> > get_w() {
        std::vector< std::vector<%(float_prec)s> > res;
        for(auto it =post_ranks.begin(); it != post_ranks.end(); it++ ) {
            res.push_back(std::move(get_dendrite_w(*it)));
        }
        return res;
    }
""",
    'init': """
""",
    'pyx_struct': """
        # Initialization
        void set_w_csr(vector[%(float_prec)s])

        # Interface access
        vector[%(float_prec)s] get_dendrite_w(int)
        vector[vector[%(float_prec)s]] get_w()
""",
    'pyx_wrapper_args': "",
    'pyx_wrapper_init': """
        proj%(id_proj)s.set_w_csr(syn._matrix.values())
""",
    'pyx_wrapper_accessor': """
    def get_w(self):
        return proj%(id_proj)s.get_w()
    def get_dendrite_w(self, int rank):
        return proj%(id_proj)s.get_dendrite_w(rank)
"""
}

single_weight_matrix = {
    'declare': """
    // Single weight in the projection
    // TODO:
""",
    'accessor': "",
    'init': "",
    'pyx_struct': """
        # Local variable w
        # TODO:
""",
    'pyx_wrapper_args': "",
    'pyx_wrapper_init': """
        # Use only the first weight
        # TODO:
""",
    'pyx_wrapper_accessor': """
    # Local variable w
    # TODO:
"""
}

inverse_connectivity_matrix = {
    'declare': """
    // CSR inverse
    std::vector<int> _col_ptr;
    std::vector<int> _row_idx;
    std::vector<int> _inv_idx;
    bool _inv_computed = false;
""",
    'init': """
        if ( _inv_computed )
            return;

        if (_row_ptr.empty()) {
            std::cerr << "no row_ptr data ..." << std::endl;
            return;
        }

        //
        // 2-pass algorithm: 1st we compute the inverse connectivity as LIL, 2ndly transform it to CSR
        //
        std::map< int, std::vector< int > > inv_post_rank = std::map< int, std::vector< int > >();
        std::map< int, std::vector< int > > inv_idx = std::map< int, std::vector< int > >();

        // iterate over post neurons, post_rank_it encodes the current rank
        for( int i = 0; i < (_row_ptr.size()-1); i++ ) {
            int row_begin = _row_ptr[i];
            int row_end = _row_ptr[i+1];

            // iterate over synapses, update both result containers
            for( int syn_idx = row_begin; syn_idx < row_end; syn_idx++ ) {
                inv_post_rank[_col_idx[syn_idx]].push_back(i);
                inv_idx[_col_idx[syn_idx]].push_back(syn_idx);
            }
        }

        //
        // store as CSR
        //
        _col_ptr.clear();
        _row_idx.clear();
        _inv_idx.clear();
        int curr_off = 0;

        // iterate over pre-neurons
        for ( int i = 0; i < %(pre_size)s; i++) {
            _col_ptr.push_back( curr_off );
            if ( !inv_post_rank[i].empty() ) {
                _row_idx.insert(_row_idx.end(), inv_post_rank[i].begin(), inv_post_rank[i].end());
                _inv_idx.insert(_inv_idx.end(), inv_idx[i].begin(), inv_idx[i].end());

                curr_off += inv_post_rank[i].size();
            }
        }
        _col_ptr.push_back(curr_off);

        if ( _nb_synapses != curr_off )
            std::cerr << "Something went wrong: (nb_synapes = " << _nb_synapses << ") != (curr_off = " << curr_off << ")" << std::endl;
    #ifdef _DEBUG_CONN
        std::cout << "Pre to Post:" << std::endl;
        for ( int i = 0; i < pop%(id_pre)s.size; i++ ) {
            std::cout << i << ": " << col_ptr[i] << " -> " << col_ptr[i+1] << std::endl;
        }
    #endif

        _inv_computed = true;
"""
}

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_acc = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector<std::vector< %(type)s > > get_%(name)s() {
        std::vector< std::vector< %(type)s > > res;
        for(auto it = post_ranks.begin(); it != post_ranks.end(); it++ ) {
            res.push_back(std::move(get_dendrite_%(name)s(*it)));
        }
        return res;
    }
    std::vector<%(type)s> get_dendrite_%(name)s(int rk) {
        std::vector<%(type)s> res;
        for(int j = _row_ptr[rk]; j < _row_ptr[rk+1]; j++)
            res.push_back(%(name)s[j]);
        return res;
    }
    %(type)s get_synapse_%(name)s(int rk_post, int rk_pre) {
        for(int j = _col_ptr[rk_post]; j < _col_ptr[rk_post+1]; j++)
            if ( _row_idx[j] == rk_pre )
                return %(name)s[_inv_idx[j]];
    }
    void set_%(name)s(std::vector<std::vector< %(type)s > >value) { }
    void set_dendrite_%(name)s(int rk, std::vector<%(type)s> value) { }
    void set_synapse_%(name)s(int rk_post, int rk_pre, %(type)s value) { }
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector<%(type)s> get_%(name)s() { return %(name)s; }
    %(type)s get_dendrite_%(name)s(int rk) { return %(name)s[rk]; }
    void set_%(name)s(std::vector<%(type)s> value) { %(name)s = value; }
    void set_dendrite_%(name)s(int rk, %(type)s value) { %(name)s[rk] = value; }
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s get_%(name)s() { return %(name)s; }
    void set_%(name)s(%(type)s value) { %(name)s = value; }
"""
}

attribute_acc_inv = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector<std::vector< %(type)s > > get_%(name)s() {
        std::vector< std::vector< %(type)s > > res;
        for(auto it = post_ranks.begin(); it != post_ranks.end(); it++ ) {
            res.push_back(std::move(get_dendrite_%(name)s(*it)));
        }
        return res;
    }
    std::vector<%(type)s> get_dendrite_%(name)s(int rk) {
        std::vector<%(type)s> res;
        for(int j = _col_ptr[rk]; j < _col_ptr[rk+1]; j++)
            res.push_back(%(name)s[_inv_idx[j]]);
        return res;
    }
    %(type)s get_synapse_%(name)s(int rk_post, int rk_pre) {
        for(int j = _col_ptr[rk_post]; j < _col_ptr[rk_post+1]; j++)
            if ( _row_idx[j] == rk_pre )
                return %(name)s[_inv_idx[j]];
    }
    void set_%(name)s(std::vector<std::vector< %(type)s > >value) { }
    void set_dendrite_%(name)s(int rk, std::vector<%(type)s> value) { }
    void set_synapse_%(name)s(int rk_post, int rk_pre, %(type)s value) { }
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector<%(type)s> get_%(name)s() { return %(name)s; }
    %(type)s get_dendrite_%(name)s(int rk) { return %(name)s[rk]; }
    void set_%(name)s(std::vector<%(type)s> value) { %(name)s = value; }
    void set_dendrite_%(name)s(int rk, %(type)s value) { %(name)s[rk] = value; }
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s get_%(name)s() { return %(name)s; }
    void set_%(name)s(%(type)s value) { %(name)s = value; }
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = std::vector< %(type)s >( _nb_synapses, %(init)s);
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = std::vector<%(type)s>( post_ranks.size(), %(init)s);
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

delay = {
    'uniform': {
        'declare': """
    // Uniform delay
    int delay ;""",
        'pyx_struct':
"""
        # Uniform delay
        int delay""",
        'init': "",
        'pyx_wrapper_init':
"""
        proj%(id_proj)s.delay = syn.uniform_delay""",
        'pyx_wrapper_accessor':
"""
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay
    def set_delay(self, value):
        proj%(id_proj)s.delay = value
"""},
    'nonuniform': {
        'declare': """
    // Non-uniform delay
    std::vector< std::vector< int > > delay ;""",
        'pyx_struct':
"""
        # Non-uniform delay
        vector[vector[int]] delay""",
        'init': "",
        'pyx_wrapper_init':
"""
        proj%(id_proj)s.delay = syn.delay""",
        'pyx_wrapper_accessor':
"""
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay[idx]
    def set_delay(self, value):
        proj%(id_proj)s.delay = value
"""
    }
}

event_driven = {
    'declare': """
    std::vector< long > _last_event;
""",
    'cpp_init': """
""",
    'pyx_struct': """
        vector[long] _last_event
""",
    'pyx_wrapper_init':
"""
        proj%(id_proj)s._last_event = vector[long]( syn._matrix.num_elements(), -10000)
"""
}

conn_templates = {
    # connectivity
    'connectivity_matrix': connectivity_matrix,
    'inverse_connectivity_matrix': inverse_connectivity_matrix,
    'weight_matrix': weight_matrix,
    'single_weight_matrix': single_weight_matrix,

    # accessors
    'attribute_decl': attribute_decl,
    'attribute_acc': attribute_acc,
    'attribute_cpp_init': attribute_cpp_init,
    'delay': delay,
    'event_driven': event_driven
}



#   ANNarchy - IF_cond_exp
#
#   A single IF neuron with exponential, conductance-based synapses, fed by two spike sources.
# 
#   This is a reimplementation of the PyNN example:
#
#   http://www.neuralensemble.org/trac/PyNN/wiki/Examples/IF_cond_exp
#
#   authors: Helge Uelo Dinkelbach, Julien Vitay

from ANNarchy import *

# Parameters
dt = 0.1
tstop = 200.0

# Setup
setup(dt=dt)

# Input populations with predetermined spike times
spike_sourceE = SpikeSourceArray(spike_times= [float(i) for i in range(5,105,10)] )
spike_sourceI = SpikeSourceArray(spike_times= [float(i) for i in range(155,255,10)])

# Population with one IF_cond_exp neuron
ifcell = Population(1, IF_cond_exp)
ifcell.set( 
    {   'i_offset' : 0.1,    'tau_refrac' : 3.0,
        'v_thresh' : -51.0,  'tau_syn_E'  : 2.0,
        'tau_syn_I': 5.0,    'v_reset'    : -70.0,
        'e_rev_E'  : 0.,     'e_rev_I'    : -80.0 } )


# Projections
connE = Projection(spike_sourceE, ifcell, 'exc').connect_all_to_all(weights=0.006, delays=2.0)
connI = Projection(spike_sourceI, ifcell, 'inh').connect_all_to_all(weights=0.02,  delays=4.0)

# Compile the network
compile()

# Simulate
m = Monitor(ifcell, ['spike', 'v'])
simulate(tstop)
data = m.get()

# Show the result
import matplotlib.pyplot as plt
plt.plot(dt*np.arange(tstop/dt), data['v'][:, 0])
plt.xlabel('Time (ms)')
plt.ylabel('Vm (mV)')
plt.ylim([-66.0, -61.0])
plt.title('IF_cond_exp')
plt.show()




#===============================================================================
#
#     OpenMPTemplates.py
#
#     This file is part of ANNarchy.
#
#     Copyright (C) 2016-2018  Julien Vitay <julien.vitay@gmail.com>,
#     Helge Uelo Dinkelbach <helge.dinkelbach@gmail.com>
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     ANNarchy is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#===============================================================================
projection_header = """#pragma once

#include "pop%(id_pre)s.hpp"
#include "pop%(id_post)s.hpp"
%(include_additional)s
%(include_profile)s

extern PopStruct%(id_pre)s pop%(id_pre)s;
extern PopStruct%(id_post)s pop%(id_post)s;
%(struct_additional)s

/////////////////////////////////////////////////////////////////////////////
// proj%(id_proj)s: %(name_pre)s -> %(name_post)s with target %(target)s
/////////////////////////////////////////////////////////////////////////////
struct ProjStruct%(id_proj)s{
    // Number of dendrites
    int size;

    // Transmission and plasticity flags
    bool _transmission, _plasticity, _update;
    int _update_period;
    long int _update_offset;

%(declare_connectivity_matrix)s
%(declare_inverse_connectivity_matrix)s
%(declare_delay)s
%(declare_event_driven)s
%(declare_rng)s
%(declare_parameters_variables)s
%(declare_additional)s
%(declare_profile)s

    // Method called to initialize the projection
    void init_projection() {
        _transmission = true;
        _update = true;
        _plasticity = true;
        _update_period = 1;
        _update_offset = 0L;

%(init_connectivity_matrix)s

        // Inverse the connectivity matrix if spiking neurons
        inverse_connectivity_matrix();

%(init_event_driven)s
%(init_parameters_variables)s
%(init_delay)s
%(init_rng)s
%(init_additional)s
%(init_profile)s
    }

    // Spiking networks: inverse the connectivity matrix
    void inverse_connectivity_matrix() {
%(init_inverse_connectivity_matrix)s
    }

    // Spiking networks: reset the ring buffer when non-uniform
    void reset_ring_buffer() {
%(reset_ring_buffer)s
    }

    // Spiking networks: update maximum delay when non-uniform
    void update_max_delay(int d){
%(update_max_delay)s
    }

    // Computes the weighted sum of inputs or updates the conductances
    void compute_psp() {
%(psp_prefix)s
%(psp_code)s
    }

    // Draws random numbers
    void update_rng() {
%(update_rng)s
    }

    // Updates synaptic variables
    void update_synapse() {
%(update_prefix)s
%(update_variables)s
    }

    // Post-synaptic events
    void post_event() {
%(post_event_prefix)s
%(post_event)s
    }

    // Accessors for default attributes
    int get_size() { return size; }
    void set_size(int new_size) { size = new_size; }

    // Additional access methods
%(access_connectivity_matrix)s
%(access_parameters_variables)s
%(access_additional)s

    // Memory management
    long int size_in_bytes() {
        long int size_in_bytes = 0;
%(determine_size)s
        return size_in_bytes;
    }

    void clear() {
    #ifdef _DEBUG
        std::cout << "PopStruct%(id_proj)s::clear()" << std::endl;
    #endif
%(clear_container)s
    }
};
"""

# Definition for the usage of C++11 STL template random
# number generators
#
# Parameters:
#
#    rd_name:
#    rd_update:
cpp_11_rng = {
    'local': {
        'decl': """    std::vector< std::vector< %(float_prec)s > > %(rd_name)s;
    %(template)s dist_%(rd_name)s;
    """,
        'init': """
        %(rd_name)s = std::vector<%(type)s>(size, 0.0);
        dist_%(rd_name)s = %(rd_init)s;
    """,
        'update': """
                %(rd_name)s[i] = dist_%(rd_name)s(rng);
    """
    },
    'global': {
        'decl': """    %(type)s %(rd_name)s;
    %(template)s dist_%(rd_name)s;
    """,
        'init': """
        %(rd_name)s = 0.0;
        dist_%(rd_name)s = %(rd_init)s;
    """,
        'update': """
            %(rd_name)s = dist_%(rd_name)s(rng);
    """
    }
}

######################################
### Structural plasticity
######################################
# All code templates needed for structural plasticity.
structural_plasticity = {
    'header_struct': {
        'header': """
    // Structural plasticity
    int dendrite_index(int post, int pre){
        int idx = -1;
        for(int i=0; i<pre_rank[post].size(); i++){
            if(pre_rank[post][i] == pre){
                idx = i;
                break;
            }
        }
        return idx;
    }
    void addSynapse(int post, int pre, double weight, int _delay=0%(extra_args)s){
        // Find where to put the synapse
        int idx = pre_rank[post].size();
        for(int i=0; i<pre_rank[post].size(); i++){
            if(pre_rank[post][i] > pre){
                idx = i;
                break;
            }
        }

        // Update connectivty
        pre_rank[post].insert(pre_rank[post].begin() + idx, pre);
        w[post].insert(w[post].begin() + idx, weight);

        // Update additional fields
%(delay_code)s
%(add_code)s
%(spike_add)s
%(rd_add)s
    };
    void removeSynapse(int post, int idx){
        pre_rank[post].erase(pre_rank[post].begin() + idx);
        w[post].erase(w[post].begin() + idx);
%(delay_remove)s
%(add_remove)s
%(spike_remove)s
%(rd_remove)s
    };
""",
        'pruning': """
    // Pruning
    bool _pruning;
    int _pruning_period;
    long int _pruning_offset;
""",
        'creating': """
    // Creating
    bool _creating;
    int _creating_period;
    long int _creating_offset;
""",
        'spiking_addcode': """
        // Add the corresponding pair in inv_pre_rank
        int idx_post = 0;
        for(int i=0; i<post_rank.size(); i++){
            if(post_rank[i] == post){
                idx_post = i;
                break;
            }
        }
        inv_pre_rank[pre].push_back(std::pair<int, int>(idx_post, idx));
""",
        'spiking_removecode': """
        // Remove the corresponding pair in inv_pre_rank
        int pre = pre_rank[post][idx];
        for(int i=0; i<inv_pre_rank[pre].size(); i++){
            if(inv_pre_rank[pre][i].second == idx){
                inv_pre_rank[pre].erase(inv_pre_rank[pre].begin() + i);
                break;
            }
        }
"""
    },
    'pyx_struct': {
        'pruning':
"""
        # Pruning
        bool _pruning
        int _pruning_period
        long _pruning_offset
""",
        'creating':
"""
        # Creating
        bool _creating
        int _creating_period
        long _creating_offset
""",
        'func':
"""
        # Structural plasticity
        int dendrite_index(int post, int pre)
        void addSynapse(int post, int pre, double weight, int _delay%(extra_args)s)
        void removeSynapse(int post, int pre)
"""
    },
    'pyx_wrapper': {
        'pruning':
"""
    # Pruning
    def start_pruning(self, int period, long offset):
        proj%(id)s._pruning = True
        proj%(id)s._pruning_period = period
        proj%(id)s._pruning_offset = offset
    def stop_pruning(self):
        proj%(id)s._pruning = False
""",
        'creating':
"""
    # Creating
    def start_creating(self, int period, long offset):
        proj%(id)s._creating = True
        proj%(id)s._creating_period = period
        proj%(id)s._creating_offset = offset
    def stop_creating(self):
        proj%(id)s._creating = False
""",
        'func':
"""
    # Structural plasticity
    def dendrite_index(self, int post_rank, int pre_rank):
        return proj%(id)s.dendrite_index(post_rank, pre_rank)
    def add_synapse(self, int post_rank, int pre_rank, double weight, int delay%(extra_args)s):
        proj%(id)s.addSynapse(post_rank, pre_rank, weight, delay%(extra_values)s)
    def remove_synapse(self, int post_rank, int pre_rank):
        proj%(id)s.removeSynapse(post_rank, proj%(id)s.dendrite_index(post_rank, pre_rank))
"""
    }
}


######################################
### Rate-coded summation OMP
######################################
# Default LiL
lil_summation_operation = {
    'sum' : """
%(pre_copy)s
nb_post = post_rank.size();
%(omp_code)s
for(int i = 0; i < nb_post; i++) {
    sum = 0.0;
    for(int j = 0; j < pre_rank[i].size(); j++) {
        sum += %(psp)s ;
    }
    pop%(id_post)s._sum_%(target)s%(post_index)s += sum;
}
""",
    'max': """
%(pre_copy)s
nb_post = post_rank.size();
%(omp_code)s
for(int i = 0; i < nb_post; i++){
    int j = 0;
    sum = %(psp)s ;
    for(int j = 1; j < pre_rank[i].size(); j++){
        if(%(psp)s > sum){
            sum = %(psp)s ;
        }
    }
    pop%(id_post)s._sum_%(target)s%(post_index)s += sum;
}
""",
    'min': """
%(pre_copy)s
nb_post = post_rank.size();
%(omp_code)s
for(int i = 0; i < nb_post; i++){
    int j= 0;
    sum = %(psp)s ;
    for(int j = 1; j < pre_rank[i].size(); j++){
        if(%(psp)s < sum){
            sum = %(psp)s ;
        }
    }
    pop%(id_post)s._sum_%(target)s%(post_index)s += sum;
}
""",
    'mean': """
%(pre_copy)s
nb_post = post_rank.size();
%(omp_code)s
for(int i = 0; i < nb_post; i++){
    sum = 0.0 ;
    for(int j = 0; j < pre_rank[i].size(); j++){
        sum += %(psp)s ;
    }
    pop%(id_post)s._sum_%(target)s%(post_index)s += sum / (double)(pre_rank[i].size());
}
"""
}

# Compressed sparse row (CSR)
csr_summation_operation = {
    'sum' : """
%(pre_copy)s
nb_post = post_ranks.size();
%(omp_code)s
for(int i = 0; i < nb_post; i++) {
    sum = 0.0;
    for(int j = _row_ptr[i]; j < _row_ptr[i+1]; j++) {
        sum += %(psp)s ;
    }
    pop%(id_post)s._sum_%(target)s[%(post_index)s] += sum;
}
""",
    'max': """
%(pre_copy)s
nb_post = post_rank.size();
%(omp_code)s
for(int i = 0; i < nb_post; i++){
    int j = _row_ptr[i];
    sum = %(psp)s ;
    for(int j = _row_ptr[i]+1; j < _row_ptr[i+1]; j++){
        if(%(psp)s > sum){
            sum = %(psp)s ;
        }
    }
    pop%(id_post)s._sum_%(target)s[%(post_index)s] += sum;
}
""",
    'min': """
%(pre_copy)s
nb_post = post_rank.size();
%(omp_code)s
for(int i = 0; i < nb_post; i++){
    int j= _row_ptr[i];
    sum = %(psp)s ;
    for(int j = _row_ptr[i]+1; j < _row_ptr[i+1]; j++){
        if(%(psp)s < sum){
            sum = %(psp)s ;
        }
    }
    pop%(id_post)s._sum_%(target)s[%(post_index)s] += sum;
}
""",
    'mean': """
%(pre_copy)s
nb_post = post_rank.size();
%(omp_code)s
for(int i = 0; i < nb_post; i++){
    sum = 0.0 ;
    for(int j = _row_ptr[i]; j < _row_ptr[i+1]; j++){
        sum += %(psp)s ;
    }
    pop%(id_post)s._sum_%(target)s[%(post_index)s] += sum / (double)(pre_rank[i].size());
}
"""
}
    
# Dense matrix
dense_summation_operation = {
    'sum' : """
%(pre_copy)s
%(omp_code)s
for(int i = 0; i < pop%(id_post)s.size; i++) {
    sum = 0.0;
    for(int j = 0; j < pop%(id_pre)s.size; j++) {
        sum += %(psp)s ;
    }
    pop%(id_post)s._sum_%(target)s[i] += sum;
}
""",
    'max': """
%(pre_copy)s
%(omp_code)s
for(int i = 0; i < pop%(id_post)s.size; i++){
    int j = 0;
    sum = %(psp)s ;
    for(int j = 1; j < pop%(id_pre)s.size; j++){
        if(%(psp)s > sum){
            sum = %(psp)s ;
        }
    }
    pop%(id_post)s._sum_%(target)s[i] += sum;
}
""",
    'min': """
%(pre_copy)s
%(omp_code)s
for(int i = 0; i < pop%(id_post)s.size; i++){
    int j= 0;
    sum = %(psp)s ;
    for(int j = 1; j < pop%(id_pre)s.size; j++){
        if(%(psp)s < sum){
            sum = %(psp)s ;
        }
    }
    pop%(id_post)s._sum_%(target)s[i] += sum;
}
""",
    'mean': """
%(pre_copy)s
%(omp_code)s
for(int i = 0; i < pop%(id_post)s.size; i++){
    sum = 0.0 ;
    for(int j = 0; j < pop%(id_pre)s.size; j++){
        sum += %(psp)s ;
    }
    pop%(id_post)s._sum_%(target)s[i] += sum / (double)(pop%(id_pre)s.size);
}
"""
}

######################################
### Spiking summation
######################################
spiking_summation_fixed_delay = """
// Event-based summation
if (_transmission && pop%(id_post)s._active){
    %(spiked_array_fusion)s

    // Iterate over all incoming spikes (possibly delayed constantly)
    %(omp_outer_loop)s
    for(int _idx_j = 0; _idx_j < %(pre_array)s.size(); _idx_j++){
        // Rank of the presynaptic neuron
        int rk_j = %(pre_array)s[_idx_j];
        // Find the presynaptic neuron in the inverse connectivity matrix
        auto inv_post_ptr = inv_pre_rank.find(rk_j);
        if (inv_post_ptr == inv_pre_rank.end())
            continue;
        // List of postsynaptic neurons receiving spikes from that neuron
        std::vector< std::pair<int, int> >& inv_post = inv_post_ptr->second;
        // Number of post neurons
        int nb_post = inv_post.size();

        %(omp_inner_loop)s
        // Iterate over connected post neurons
        for(int _idx_i = 0; _idx_i < nb_post; _idx_i++){
            // Retrieve the correct indices
            int i = inv_post[_idx_i].first;
            int j = inv_post[_idx_i].second;

            // Event-driven integration
            %(event_driven)s
            // Update conductance
            %(g_target)s
            // Synaptic plasticity: pre-events
            %(pre_event)s
        }
    }

    %(omp_reduce_code)s
} // active
"""

spiking_summation_fixed_delay_csr ={
    'post_to_pre': """// Event-based summation
if (_transmission && pop%(id_post)s._active){

    // Iterate over all spiking neurons
    %(omp_code)s
    for( int _idx = 0; _idx < %(pre_array)s.size(); _idx++) {
        int _pre = %(pre_array)s[_idx];
    #ifdef _OPENMP
        int thr = omp_get_thread_num();
    #endif
        // Iterate over connected post neurons
        for(int syn = _col_ptr[_pre]; syn < _col_ptr[_pre + 1]; syn++) {
            %(event_driven)s
            %(g_target)s
            %(pre_event)s
        }
    }
%(omp_reduce_code)s
} // active
""",
    'pre_to_post': """// Event-based summation
if (_transmission && pop%(id_post)s._active){

    %(omp_code)s
    // Iterate over all spiking neurons
    for( int _idx = 0; _idx < %(pre_array)s.size(); _idx++) {
        // Rank of the presynaptic neuron
        int _pre = %(pre_array)s[_idx];

        // Iterate over connected post neurons
        for(int syn = _row_ptr[_pre]; syn < _row_ptr[_pre + 1]; syn++) {

            // Event-driven integration
            %(event_driven)s
            // Update conductance
            %(g_target)s
            // Synaptic plasticity: pre-events
            %(pre_event)s
        }
    }
} // active
"""
}

spiking_summation_fixed_delay_dense_matrix = """
// Event-based summation
if (_transmission && pop%(id_post)s._active){
    // TODO?
} // active
"""

# Uses a ring buffer to process non-uniform delays in spiking networks
spiking_summation_variable_delay = """
// Event-based summation
if (_transmission && pop%(id_post)s._active){

    // Iterate over the spikes emitted during the last step in the pre population
    for(int idx_spike=0; idx_spike<pop%(id_pre)s.spiked.size(); idx_spike++){

        // Get the rank of the pre-synaptic neuron which spiked
        int rk_pre = pop%(id_pre)s.spiked[idx_spike];
        // List of post neurons receiving connections
        std::vector< std::pair<int, int> > rks_post = inv_pre_rank[rk_pre];

        // Iterate over the post neurons
        for(int x=0; x<rks_post.size(); x++){
            // Index of the post neuron in the connectivity matrix
            int i = rks_post[x].first ;
            // Index of the pre neuron in the connecivity matrix
            int j = rks_post[x].second ;
            // Delay of that connection
            int d = delay[i][j]-1;
            // Index in the ring buffer
            int modulo_delay = (idx_delay + d) %% max_delay;
            // Add the spike in the ring buffer
            _delayed_spikes[modulo_delay][i].push_back(j);
        }
    }

    // Iterate over all post neurons having received spikes in the previous steps
    for (int i=0; i<_delayed_spikes[idx_delay].size(); i++){
        for (int _idx_j=0; _idx_j<_delayed_spikes[idx_delay][i].size(); _idx_j++){
            // Pre-synaptic index in the connectivity matrix
            int j = _delayed_spikes[idx_delay][i][_idx_j];

            // Event-driven integration
            %(event_driven)s
            // Update conductance
            %(g_target)s
            // Synaptic plasticity: pre-events
            %(pre_event)s
        }
        // Empty the current list of the ring buffer
        _delayed_spikes[idx_delay][i].clear();
    }

    // Increment the index of the ring buffer
    idx_delay = (idx_delay + 1) %% max_delay;

} // active
"""

"""
    // Old stuff just in case
    // Iterate over all post neurons
    //%(omp_code)s
    for (int i=0; i<post_rank.size(); i++){
        for (int j=0; j<pre_rank[i].size(); j++){
            int d = delay[i][j]-1;
            if(std::find(pop%(id_pre)s._delayed_spike[d].begin(), pop%(id_pre)s._delayed_spike[d].end(), pre_rank[i][j]) != pop%(id_pre)s._delayed_spike[d].end()){

                %(event_driven)s
                %(g_target)s
                %(pre_event)s
            }
        }
    }
"""

######################################
### Update synaptic variables
######################################
lil_update_variables = {
    'local': """
// Check periodicity
if(_transmission && _update && pop%(id_post)s._active && ( (t - _update_offset)%%_update_period == 0L) ){
    // Global variables
    %(global)s
    // Local variables
    %(omp_code)s
    for(int i = 0; i < post_rank.size(); i++){
        rk_post = post_rank[i]; // Get postsynaptic rank
        // Semi-global variables
        %(semiglobal)s
        // Local variables
        for(int j = 0; j < pre_rank[i].size(); j++){
            rk_pre = pre_rank[i][j]; // Get presynaptic rank
    %(local)s
        }
    }
}
""",
    'global': """
// Check periodicity
if(_transmission && _update && pop%(id_post)s._active && ( (t - _update_offset)%%_update_period == 0L)){
    // Global variables
    %(global)s
    // Local variables
    %(omp_code)s
    for(int i = 0; i < post_rank.size(); i++){
        rk_post = post_rank[i]; // Get postsynaptic rank
    %(semiglobal)s
    }
}
"""
}

csr_update_variables = {
    'local': """
if(_transmission && _update && pop%(id_post)s._active && ( (t - _update_offset)%%_update_period == 0L) ){
    %(global)s
    %(omp_code)s
    for(int i = 0; i < post_ranks.size(); i++){
        rk_post = post_ranks[i];
    %(semiglobal)s
        for(int j = _row_ptr[rk_post]; j < _row_ptr[rk_post+1]; j++){
            rk_pre = _col_idx[j];
    %(local)s
        }
    }
}
""",
    'global': """
if(_transmission && _update && pop%(id_post)s._active && ( (t - _update_offset)%%_update_period == 0L)){
    %(global)s
    %(omp_code)s
    for(int i = 0; i < post_ranks.size(); i++){
        rk_post = post_ranks[i];
    %(semiglobal)s
    }
}
"""
}

dense_update_variables = {
    'local': """
// Check periodicity
if(_transmission && _update && pop%(id_post)s._active && ( (t - _update_offset)%%_update_period == 0L)){
    // Global variables
    %(global)s
    // Local variables
    %(omp_code)s
    for(int i = 0; i < pop%(id_post)s.size; i++){
        rk_post = i; // dense: ranks are indices
        // Semi-global variables
    %(semiglobal)s
        for(int j = 0; j < pop%(id_pre)s.size; j++){
            rk_pre = j; // dense: ranks are indices
    %(local)s
        }
    }
}
""",
    'global': """
// Check periodicity
if(_transmission && _update && pop%(id_post)s._active && ( (t - _update_offset)%%_update_period == 0L)){
    // Global variables
    %(global)s
    // Semi-global variables
    %(omp_code)s
    for(int i = 0; i < pop%(id_post)s.size; i++){
        rk_post = i;
    %(semiglobal)s
    }
}
"""
}

openmp_templates = {
    'projection_header': projection_header,
    'rng': cpp_11_rng
}



# Generic imports
from __future__ import print_function
import os, sys
import numpy as np

# ANNarchy core
from .core.Global import *
from .core.Simulate import *
from .core.Neuron import Neuron
from .core.Synapse import Synapse
from .core.Population import Population
from .core.Projection import Projection
from .core.SpecificPopulation import PoissonPopulation, SpikeSourceArray, TimedArray, HomogeneousCorrelatedSpikeTrains
from .core.SpecificProjection import DecodingProjection, CurrentInjection
from .core.Dendrite import Dendrite
from .core.Random import Uniform, DiscreteUniform, Normal, LogNormal, Gamma, Exponential
from .core.IO import save, load, load_parameter, load_parameters, save_parameters
from .core.Utils import sparse_random_matrix
from .core.Monitor import Monitor, raster_plot, histogram, population_rate, smoothed_rate, mean_fr
from .core.Network import Network, parallel_run
from .core.Profiler import Profiler
from .parser.report.Report import report
from .models import *
from .extensions import *

# Cython modules
try:
    # HD: until version 4.6 the connectivity class wasn't named properly. To ensure backward compability
    #     we rename the LILConnectivity to CSR
    from .core.cython_ext import LILConnectivity as CSR
except Exception as e:
    print(e)
    print("""
Warning: Cython modules can not be imported. If you are installing ANNarchy, this is normal, ignore this message. If ANNarchy is already installed, something went wrong with the compilation, try reinstalling.
""")

# ANNarchy compilation
from .generator import compile

# Version
__version__ = '4.6'
__release__ = '4.6.9.7'

core.Global._print( 'ANNarchy ' + __version__ + ' (' + __release__ + \
                    ') on ' + sys.platform + ' (' + os.name + ').' )



#   ANNarchy - EIF_cond_exp
#
#   Network of EIF neurons with exponentially decreasing conductance-based synapses.
# 
#   This is a reimplementation of the Brian example:
#
#   http://brian.readthedocs.org/en/1.4.1/examples-misc_expIF_network.html
#
#   authors: Helge Uelo Dinkelbach, Julien Vitay

from ANNarchy import *

dt = 0.1
setup(dt=dt)

EIF = Neuron(
    parameters = """
    v_rest = -70.0 : population
    cm = 0.2 : population
    tau_m = 10.0 : population
    tau_syn_E = 5.0 : population
    tau_syn_I = 10.0 : population
    e_rev_E = 0.0 : population
    e_rev_I = -80.0 : population
    delta_T = 3.0 : population
    v_thresh = -55.0 : population
    v_reset = -70.0 : population
    v_spike = -20.0 : population
""",
    equations="""    
    dv/dt = (v_rest - v +  delta_T * exp( (v-v_thresh)/delta_T) )/tau_m + ( g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) )/cm
    
    tau_syn_E * dg_exc/dt = - g_exc 
    tau_syn_I * dg_inh/dt = - g_inh 
""",
    spike = """
    v > v_spike
""",
    reset = """
    v = v_reset
""",
    refractory = 2.0

)

# Global population
P = Population(geometry=4000, neuron=EIF)

# Subpopulations
Pe = P[:3200]
Pi = P[3200:]
Pe_input = P[:200]
Pi_input = P[3200:3400]

# Projections
we = 1.5 / 1000.0 # excitatory synaptic weight
wi = 2.5 * we # inhibitory synaptic weight
Ce = Projection(Pe, P, 'exc').connect_fixed_probability(weights=we, probability=0.05)
Ci = Projection(Pi, P, 'inh').connect_fixed_probability(weights=wi, probability=0.05)

# Initialization of variables
P.v = -70.0 + 10.0 * np.random.rand(P.size)
P.g_exc = (np.random.randn(P.size) * 2.0 + 5.0) * we
P.g_inh = (np.random.randn(P.size) * 2.0 + 5.0) * wi

# Poisson inputs
i_exc = PoissonPopulation(geometry=200, rates="if t < 200.0 : 2000.0 else : 0.0")
i_inh = PoissonPopulation(geometry=200, rates="if t < 100.0 : 2000.0 else : 0.0")
Ie = Projection(i_exc, Pe_input, 'exc').connect_one_to_one(weights=we)
Ii = Projection(i_inh, Pi_input, 'exc').connect_one_to_one(weights=we)

# Compile the Network
compile()

# Simulate
m = Monitor(P, 'spike')
simulate(500.0, measure_time=True)

# Retrieve recordings
data = m.get()
t, n = m.raster_plot(data['spike'])

# Plot
import matplotlib.pyplot as plt
plt.plot(t, n, '.')
plt.show()



#===============================================================================
#
#     NetworkManager.py
#
#     This file is part of ANNarchy.
#
#     Copyright (C) 2020 Helge Uelo Dinkelbach <helge.dinkelbach@gmail.com>,
#     Julien Vitay <julien.vitay@gmail.com>
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     ANNarchy is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#===============================================================================
from ANNarchy.core import Global
import os

class NetworkManager(object):
    """
    This class implements the management of the data for the different networks. The
    add/remove methods allow the 'random' removal of the network instances and later
    refill of empty spaces.

    The class will be placed and instantiated in the ANNarchy.core.Global file.
    """
    def __init__(self):
        """
        Constructor.
        """
        self._create_initial_state()

    def _create_initial_state(self):
        """
        Initialize the container for the initial network.

        Called either from __init__ or clear(). The first
        slot is reserved for the magic network.
        """
        self._network = [
            {
                'populations': [],
                'projections': [],
                'monitors': [],
                'instance': None,
                'compiled': False,
                'directory': os.getcwd() + "/annarchy/"
            },
        ]
        self._py_instances = [None]

    def __getitem__(self, net_id):
        """
        In the ANNarchy framework we need to access easily the network
        configuration. This method makes the NetworkManager easily subscribtable:

            Global._network[net_id]
        """
        if isinstance(net_id, int):
            if net_id < len(self._network):
                return self._network[net_id]
            else:
                Global._error("Network", net_id, "not existing ...")
        elif isinstance(net_id, slice):
            return self._network[net_id]

    def __len__(self):
        """
        Called if len() is applied on a NetworkManager class.
        """
        return len(self._network)

    def add_network(self, py_instance):
        """
        Adds an empty structure for a new network and returns the new network ID.
        """
        new_dict = {
            'populations': [],
            'projections': [],
            'monitors': [],
            'instance': None,
            'compiled': False,
            'directory': os.getcwd() + "/annarchy/"
        }

        found = -1
        # scan for slots which were freed before
        for i, entry in enumerate(self._network):
            if entry == None and self._py_instance[i] == None:
                found = i
                break

        # dependent on the scan append at the end
        # or fill free slot
        if found == -1:
            new_id = len(self._network)
            self._network.append(new_dict)
            self._py_instances.append(py_instance)
        else:
            new_id = found
            self._network[new_id] = new_dict
            self._py_instances[new_id] = py_instance
        Global._debug("Added network", new_id)
        return new_id

    def _remove_network(self, py_instance):
        """
        Remove the given network from the list of compilable/instantiable networks.
        It is important to invalidate only the slot. If del is called on the dictionary
        entry then this will lead to a removal of the space and therefore all succesequent
        networks would be assigned wrong.

        This function will be called from the Network.__del__() method, after destruction
        of the attached objects.
        """
        Global._debug("Remove network", py_instance)

        net_id = -1
        for net_id, inst in enumerate(self._py_instances):
             if inst == py_instance:
                 self._network[net_id] = None
                 self._py_instances[net_id] = None

    def clear(self):
        """
        Remove all assigned networks and restore the initial state.
        """
        del self._network
        self._create_initial_state()

